from orchestra.bots.staffbot import StaffBot
from orchestra.models import StaffBotRequest


def staffbot_autoassign(task, **kwargs):
    request_cause = StaffBotRequest.RequestCause.AUTOSTAFF.value
    bot = StaffBot()
    bot.staff(task.id, request_cause=request_cause)
    return task



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-12 19:56
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0032_payrate'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tasktimer',
            name='assignment',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE,
                                    related_name='timers', to='orchestra.TaskAssignment'),
        ),
        migrations.AlterField(
            model_name='tasktimer',
            name='worker',
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE, related_name='timer', to='orchestra.Worker'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-28 17:51
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0049_auto_20160428_1750'),
    ]

    operations = [
        migrations.AlterField(
            model_name='staffingrequest',
            name='communication_method',
            field=models.IntegerField(choices=[(0, 'slack'), (1, 'email')]),
        ),
    ]



import json
from unittest.mock import patch

from django.conf import settings
from django.contrib.auth.models import Group
from django.core.urlresolvers import reverse
from django.test import Client
from django.test import override_settings

from orchestra.core.errors import InvalidRevertError
from orchestra.models import Project
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.project_api.serializers import IterationSerializer
from orchestra.project_api.serializers import TaskAssignmentSerializer
from orchestra.project_api.serializers import TaskSerializer
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import UserFactory
from orchestra.tests.helpers.fixtures import WorkerFactory
from orchestra.tests.helpers.fixtures import setup_complete_task
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.tests.helpers.iterations import verify_iterations
from orchestra.utils.load_json import load_encoded_json
from orchestra.utils.revert import RevertChange
from orchestra.utils.task_lifecycle import create_subsequent_tasks
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_lifecycle import complete_and_skip_task
from orchestra.utils.task_properties import assignment_history
from orchestra.utils.task_properties import current_assignment
from orchestra.utils.task_properties import get_iteration_history
from orchestra.utils.task_properties import get_latest_iteration


class ProjectManagementAPITestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.project_admin_group = Group.objects.create(name='project_admins')
        user = (UserFactory(username='project_management',
                            first_name='project_management',
                            last_name='project_management',
                            password='project_management',
                            email='project_management@b12.io'))
        self.worker = WorkerFactory(user=user)
        user.groups.add(self.project_admin_group)

        self.api_client = Client()
        self.api_client.login(
            username='project_management', password='project_management')

    def test_project_information_api(self):
        project = self.projects['project_management_project']
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:'
                'project_information'),
            json.dumps({
                'project_id': project.id,
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        returned = load_encoded_json(response.content)

        # Skipping the `tasks` key/value pair for this test
        expected_project = {
            'task_class': project.task_class,
            'start_datetime': '2015-10-12T00:00:00Z',
            'team_messages_url': None,
            'admin_url': settings.ORCHESTRA_URL + reverse(
                'admin:orchestra_project_change',
                args=(project.id,)),
            'priority': project.priority,
            'project_data': project.project_data,
            'short_description': project.short_description,
            'id': project.id,
            'workflow_slug': project.workflow_version.workflow.slug
        }
        returned_project = returned[str(project.id)]['project']
        self.assertEqual(
            expected_project,
            {k: returned_project[k] for k in expected_project.keys()})

        sample_task = returned[str(project.id)]['tasks']['step1']
        task = project.tasks.get(step__slug='step1')
        expected_task = {
            'status': 'Post-review Processing',
            'start_datetime': '2015-10-12T01:00:00Z',
            'admin_url': settings.ORCHESTRA_URL + reverse(
                'admin:orchestra_task_change',
                args=(task.id,)),
            'latest_data': {
                'test_key': 'test_value'
            },
            'project': task.project.id,
            'step_slug': 'step1',
            'id': task.id
        }

        self.assertEqual(
            expected_task,
            {k: sample_task[k] for k in expected_task.keys()})

        sample_assignment = sample_task['assignments'][0]
        assignment = assignment_history(task)[0]
        expected_assignment = {
            'status': 'Submitted',
            'start_datetime': '2015-10-12T02:00:00Z',
            'task': assignment.task.id,
            'admin_url': settings.ORCHESTRA_URL + reverse(
                'admin:orchestra_taskassignment_change',
                args=(assignment.id,)),
            'worker': {
                'username': assignment.worker.user.username,
                'first_name': assignment.worker.user.first_name,
                'last_name': assignment.worker.user.last_name,
                'id': assignment.worker.id,
                'slack_username': assignment.worker.slack_username,
                'slack_user_id': assignment.worker.slack_user_id
            },
            'iterations': [
                {
                    'id': assignment.iterations.first().id,
                    'admin_url': settings.ORCHESTRA_URL + reverse(
                        'admin:orchestra_iteration_change',
                        args=(assignment.iterations.first().id,)),
                    'assignment': assignment.id,
                    'start_datetime': '2015-10-12T02:00:00Z',
                    'end_datetime': '2015-10-12T03:00:00Z',
                    'status': 'Requested Review',
                    'submitted_data': {}
                }
            ],
            'in_progress_task_data': {},
            'id': assignment.id
        }

        self.assertEqual(
            expected_assignment,
            {k: sample_assignment[k] for k in expected_assignment.keys()})

    def test_reassign_assignment_api(self):
        task = self.tasks['project_management_task']
        # Reassign entry-level assignment to entry-level worker
        entry_assignment = task.assignments.get(assignment_counter=0)
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:reassign_assignment'),
            json.dumps({
                'worker_username': self.workers[0].user.username,
                'assignment_id': entry_assignment.id
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        entry_assignment.refresh_from_db()
        self.assertEqual(entry_assignment.worker, self.workers[0])

        # Reassign entry-level assignment to reviewer
        entry_assignment.refresh_from_db()
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:reassign_assignment'),
            json.dumps({
                'worker_username': self.workers[1].user.username,
                'assignment_id': entry_assignment.id
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        entry_assignment.refresh_from_db()
        self.assertEqual(entry_assignment.worker, self.workers[1])

        # Attempt to reassign assignment to worker already working on task
        first_review_assignment = task.assignments.get(assignment_counter=1)
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:reassign_assignment'),
            json.dumps({
                'worker_username': self.workers[1].user.username,
                'assignment_id': first_review_assignment.id
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        first_review_assignment.refresh_from_db()
        self.assertTrue(first_review_assignment.worker, self.workers[1])
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Worker already assigned to this task.')

        # Attempt to reassign review assignment to entry-level worker
        first_review_assignment = task.assignments.get(assignment_counter=1)
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:reassign_assignment'),
            json.dumps({
                'worker_username': self.workers[0].user.username,
                'assignment_id': first_review_assignment.id
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        first_review_assignment.refresh_from_db()
        self.assertTrue(first_review_assignment.worker, self.workers[0])
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Worker not certified for this assignment.')

        # Reassign review assignment to another reviewer
        second_review_assignment = task.assignments.get(assignment_counter=2)
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:reassign_assignment'),
            json.dumps({
                'worker_username': self.workers[3].user.username,
                'assignment_id': second_review_assignment.id
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        second_review_assignment.refresh_from_db()
        self.assertEqual(second_review_assignment.worker, self.workers[3])

    def test_complete_and_skip_task_api(self):
        # This task, when submitted, will be moved into reviewing
        # state because it's already been assigned.
        task = self.tasks['project_management_task']
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:'
                'complete_and_skip_task'),
            json.dumps({
                'task_id': task.id,
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        task.refresh_from_db()
        self.assertEqual(task.status, Task.Status.REVIEWING)
        assignment_statuses = [
            TaskAssignment.Status.SUBMITTED, TaskAssignment.Status.SUBMITTED,
            TaskAssignment.Status.PROCESSING]
        for index, assignment in enumerate(
                task.assignments.all().order_by('assignment_counter')):
            self.assertEqual(
                assignment.status, assignment_statuses[index])
        # Check that dependent tasks have already been created
        num_tasks = task.project.tasks.count()
        create_subsequent_tasks(task.project)
        self.assertEqual(task.project.tasks.count(), num_tasks)

        # This task, when submitted, will be moved into a completed
        # state because it has no assignments.
        task = self.tasks['awaiting_processing']
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:'
                'complete_and_skip_task'),
            json.dumps({
                'task_id': task.id,
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        task.refresh_from_db()
        self.assertEqual(task.status, Task.Status.COMPLETE)

    @override_settings(ORCHESTRA_SLACK_EXPERTS_ENABLED=True)
    @patch('orchestra.communication.tests.helpers.slack.Groups.unarchive')
    @patch('orchestra.communication.tests.helpers.slack.Groups.archive')
    def test_unarchive_slack_channel_api(self, mock_archive, mock_unarchive):
        project = self.projects['single_human_step']
        create_subsequent_tasks(project)
        task = Task.objects.get(project=project)
        assign_task(self.workers[1].id, task.id)
        complete_and_skip_task(task.id)
        create_subsequent_tasks(project)
        self.assertTrue(mock_archive.called)

        response = self.api_client.post(
            reverse(('orchestra:orchestra:project_management'
                     ':unarchive_slack_channel')),
            json.dumps({
                'project_id': project.id,
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertTrue(mock_unarchive.called)

    def test_end_project_api(self):
        project = self.projects['project_to_end']
        response = self.api_client.post(
            reverse('orchestra:orchestra:project_management:end_project'),
            json.dumps({
                'project_id': project.id,
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        project.refresh_from_db()
        self.assertEqual(project.status, Project.Status.ABORTED)
        for task in Task.objects.filter(project=project):
            self.assertEqual(task.status, Task.Status.ABORTED)

    def test_set_project_status_api(self):
        project = self.projects['project_to_set_status']
        status_choices = dict(Project.STATUS_CHOICES)
        paused_status = status_choices[Project.Status.PAUSED]
        active_status = status_choices[Project.Status.ACTIVE]
        response = self.api_client.post(
            reverse((
                'orchestra:orchestra:project_management:'
                'set_project_status')),
            json.dumps({
                'project_id': project.id,
                'status': paused_status
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        project.refresh_from_db()
        self.assertEqual(project.status, Project.Status.PAUSED)

        response = self.api_client.post(
            reverse((
                'orchestra:orchestra:project_management:'
                'set_project_status')),
            json.dumps({
                'project_id': project.id,
                'status': active_status
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        project.refresh_from_db()
        self.assertEqual(project.status, Project.Status.ACTIVE)

    def test_invalid_revert_aborted_task(self):
        task = setup_complete_task(self)
        task.status = Task.Status.ABORTED
        task.save()

        with self.assertRaises(InvalidRevertError):
            self._revert_task(
                task, task.assignments.first().iterations.first(),
                revert_before=False, commit=True)

        task.delete()

    def test_invalid_revert_before(self):
        task = setup_complete_task(self)
        task.status = Task.Status.ABORTED
        task.save()

        # Ensure reviewer assignment has more than one iteration
        reviewer_assignment = assignment_history(task).last()
        self.assertGreater(reviewer_assignment.iterations.count(), 1)

        # Attempt to revert before an iteration that isn't the assignment's
        # first
        with self.assertRaises(InvalidRevertError):
            self._revert_task(
                task, get_latest_iteration(reviewer_assignment),
                revert_before=True, commit=True)

        task.delete()

    def test_revert_second_review(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.REVIEWING

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.UNCHANGED.value, RevertChange.REVERTED.value),
            iteration_changes=(
                (RevertChange.UNCHANGED.value, RevertChange.UNCHANGED.value),
                (RevertChange.UNCHANGED.value, RevertChange.REVERTED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[3],
            num_iterations=4,
            task_status=reverted_status,
            latest_data={'test': 'reviewer_accept'},
            expected_audit=expected_audit)

        task.delete()

    def test_revert_post_review_processing(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.POST_REVIEW_PROCESSING

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.REVERTED.value, RevertChange.REVERTED.value),
            iteration_changes=(
                (RevertChange.UNCHANGED.value, RevertChange.REVERTED.value),
                (RevertChange.UNCHANGED.value, RevertChange.DELETED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[2],
            num_iterations=3,
            task_status=reverted_status,
            latest_data={'test': 'entry_resubmit'},
            expected_audit=expected_audit)

        task.delete()

    def test_revert_first_review(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.REVIEWING

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.REVERTED.value, RevertChange.REVERTED.value),
            iteration_changes=(
                (RevertChange.UNCHANGED.value, RevertChange.DELETED.value),
                (RevertChange.REVERTED.value, RevertChange.DELETED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[1],
            num_iterations=2,
            task_status=reverted_status,
            latest_data={'test': 'reviewer_accept'},
            expected_audit=expected_audit)

        task.delete()

    def test_revert_pending_review(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.PENDING_REVIEW

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.REVERTED.value, RevertChange.DELETED.value),
            iteration_changes=(
                (RevertChange.UNCHANGED.value, RevertChange.DELETED.value),
                (RevertChange.DELETED.value, RevertChange.DELETED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[1],
            num_iterations=1,
            task_status=reverted_status,
            latest_data={'test': 'entry_resubmit'},
            expected_audit=expected_audit,
            revert_before=True)

        task.delete()

    def test_revert_processing(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.PROCESSING

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.REVERTED.value, RevertChange.DELETED.value),
            iteration_changes=(
                (RevertChange.REVERTED.value, RevertChange.DELETED.value),
                (RevertChange.DELETED.value, RevertChange.DELETED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[0],
            num_iterations=1,
            task_status=reverted_status,
            latest_data={'test': 'entry_resubmit'},
            expected_audit=expected_audit)

        task.delete()

    def test_revert_awaiting_processing(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.AWAITING_PROCESSING

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.DELETED.value, RevertChange.DELETED.value),
            iteration_changes=(
                (RevertChange.DELETED.value, RevertChange.DELETED.value),
                (RevertChange.DELETED.value, RevertChange.DELETED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[0],
            num_iterations=0,
            task_status=reverted_status,
            latest_data={'test': 'entry_resubmit'},
            expected_audit=expected_audit,
            revert_before=True)

        task.delete()

    def _revert_task(self, task, iteration, revert_before, commit):
        return self.api_client.post(
            reverse('orchestra:orchestra:project_management:revert_task'),
            json.dumps({
                'task_id': task.id,
                # Convert datetime to timestamp
                'iteration_id': iteration.id,
                'revert_before': revert_before,
                'commit': commit,
            }),
            content_type='application/json')

    def _test_reverted_task(self, task, iteration, num_iterations,
                            task_status, latest_data, expected_audit,
                            revert_before=False):
        response = self._revert_task(
            task, iteration, revert_before=revert_before, commit=False)
        self.assertEqual(response.status_code, 200)
        task.refresh_from_db()
        fake_audit = load_encoded_json(response.content)
        self.assertEqual(fake_audit, expected_audit)

        self.assertEqual(task.status, Task.Status.COMPLETE)
        self.assertEqual(task.assignments.count(), 2)
        for assignment in task.assignments.all():
            self.assertEqual(assignment.iterations.count(), 2)

        response = self._revert_task(
            task, iteration, revert_before=revert_before, commit=True)
        self.assertEqual(response.status_code, 200)
        task.refresh_from_db()
        audit = load_encoded_json(response.content)
        self.assertEqual(audit, fake_audit)
        task.refresh_from_db()
        self.assertEqual(task.status, task_status)
        self.assertEqual(
            get_iteration_history(task).count(), num_iterations)

        verify_iterations(task.id)

        if num_iterations:
            self.assertEqual(
                current_assignment(task).in_progress_task_data,
                latest_data)

        verify_iterations(task.id)

    def _expected_audit(self, complete_task, reverted_status=None,
                        assignment_changes=None, iteration_changes=None):
        assignments = assignment_history(complete_task)
        iterations = get_iteration_history(complete_task)
        audit = {
            'task': TaskSerializer(complete_task).data,
            'assignments': [
                {
                    'assignment': (
                        TaskAssignmentSerializer(assignments.first()).data),
                    'change': 0,
                    'iterations': [
                        {
                            'change': 0,
                            'iteration': (
                                IterationSerializer(iterations.all()[0]).data)
                        },
                        {
                            'change': 0,
                            'iteration': (
                                IterationSerializer(iterations.all()[2]).data)
                        }
                    ]
                },
                {
                    'assignment': (
                        TaskAssignmentSerializer(assignments.last()).data),
                    'change': 0,
                    'iterations': [
                        {
                            'change': 0,
                            'iteration': (
                                IterationSerializer(iterations.all()[1]).data)
                        },
                        {
                            'change': 0,
                            'iteration': (
                                IterationSerializer(iterations.all()[3]).data)
                        }
                    ]
                }
            ],
        }
        if reverted_status is not None:
            audit['reverted_status'] = reverted_status
        if assignment_changes is not None:
            for i, assignment_change in enumerate(assignment_changes):
                audit['assignments'][i]['change'] = assignment_change
        if iteration_changes is not None:
            for i, changes_per_assignment in enumerate(iteration_changes):
                for j, iteration_change in enumerate(changes_per_assignment):
                    audit['assignments'][i][
                        'iterations'][j]['change'] = iteration_change
        return audit



from django.core.management.base import BaseCommand

from orchestra.workflow.certifications import migrate_certifications


class Command(BaseCommand):
    help = 'Migrates specified certifications across workflows.'

    def add_arguments(self, parser):
        parser.add_argument(
            'source_workflow_slug',
            help='Slug for the source workflow from which to migrate '
                 'certifications.')
        parser.add_argument(
            'destination_workflow_slug',
            help='Slug for the destination workflow to which certifications '
                 'will be migrated.')
        parser.add_argument(
            '--certifications',
            nargs='+',
            metavar=('slug1', 'slug2'),
            help=('Certification slugs to migrate. If not specified, '
                  'migration of all certifications will be attempted.'))

    def handle(self, *args, **options):
        migrate_certifications(
            options['source_workflow_slug'],
            options['destination_workflow_slug'],
            options['certifications'])



from bitfield import BitField
from django.db import models

from orchestra.models.communication.managers import \
    CommunicationPreferenceManager
from orchestra.models.communication.mixins import CommunicationPreferenceMixin
from orchestra.models.communication.mixins import StaffBotRequestMixin
from orchestra.models.communication.mixins import StaffingRequestInquiryMixin
from orchestra.models.communication.mixins import StaffingResponseMixin
from orchestra.models.core.models import Task
from orchestra.models.core.models import Worker
from orchestra.utils.models import BaseModel
from orchestra.utils.models import ChoicesEnum


class CommunicationPreference(CommunicationPreferenceMixin, BaseModel):
    """
        A CommunicationPreference object defines how a Worker would like to
        contacted for a given CommunicationType.

        Attributes:
            worker (orchestra.models.Worker):
                Django user that the preference represents.
            methods (BitField):
                The ways in which the user would like to be contacted.
            type (CommunicationType):
                The type of communication to which this preference applies.
    """
    objects = CommunicationPreferenceManager()

    class CommunicationMethods:
        SLACK = 'slack'
        EMAIL = 'email'

    COMMUNICATION_METHODS = (
        (CommunicationMethods.SLACK, 'Slack'),
        (CommunicationMethods.EMAIL, 'Email'),
    )

    class CommunicationType(ChoicesEnum):
        TASK_STATUS_CHANGE = 'task_status_change'
        NEW_TASK_AVAILABLE = 'new_task_available'

    COMMUNICATION_TYPE_DESCRIPTIONS = {
        CommunicationType.TASK_STATUS_CHANGE:
        {
            'short_description': 'Task Status Changes',
            'long_description':
            """
            When a task status changes (e.g., when it moves from in
            review to returned from reviewer), you will automatically
            receive a notification in the project Slack group. Select
            whether you would like to receive an email notification as
            well.
            """
        },
        CommunicationType.NEW_TASK_AVAILABLE:
        {
            'short_description': 'New Tasks',
            'long_description':
            """
            Select your preferred channel for being notified when a
            new task is available to work on.
            """
        }
    }

    worker = models.ForeignKey(Worker, on_delete=models.CASCADE)
    methods = BitField(flags=COMMUNICATION_METHODS,
                       blank=True, null=True, default=None)
    communication_type = models.IntegerField(
        choices=CommunicationType.choices())

    class Meta:
        app_label = 'orchestra'
        unique_together = (('worker', 'communication_type'),)


class StaffBotRequest(StaffBotRequestMixin, BaseModel):
    """
    A StaffBotRequest object defines a new task that needs
    a worker for a specific role.

    Attributes:
        task (orchestra.models.Task):
            The task that needs a worker assignment
        required_role_counter (int):
            Counter that tells which expert is a new worker
            going to be on a task
        request_cause (RequestCause):
            The cause for request
        project_description (str):
            Description of the project
        status (Status)
            Status of the request
        last_inquiry_sent (DateTimeField):
            The datetime the most recent inquiry was sent.
    """
    class RequestCause(ChoicesEnum):
        USER = 'user'
        AUTOSTAFF = 'autostaff'
        RESTAFF = 'restaff'

    class Status(ChoicesEnum):
        PROCESSING = 'processing'
        COMPLETE = 'complete'

    task = models.ForeignKey(
        Task, related_name='staffing_requests', on_delete=models.CASCADE)
    required_role_counter = models.IntegerField()
    request_cause = models.IntegerField(choices=RequestCause.choices())
    project_description = models.TextField(null=True, blank=True)
    status = models.IntegerField(
        default=Status.PROCESSING.value,
        choices=Status.choices())
    last_inquiry_sent = models.DateTimeField(null=True)


class StaffingRequestInquiry(StaffingRequestInquiryMixin, BaseModel):
    """
    A StaffingRequestInquiry object defines how a Worker was contacted to
    work on a StaffBotRequest

    Attributes:
        communication_preference (orchestra.models.CommunicationPreference):
            Django user that the request is sent to
        request (orchestra.models.StaffBotRequest):
            Request object associated with inquiry
        communication_method (CommunicationMethod):
            Method by which a worker is going to be contacted
    """

    class CommunicationMethod(ChoicesEnum):
        SLACK = 'slack'
        EMAIL = 'email'

    request = models.ForeignKey(StaffBotRequest,
                                on_delete=models.CASCADE,
                                null=True,
                                related_name='inquiries')
    communication_preference = models.ForeignKey(
        CommunicationPreference, on_delete=models.CASCADE)
    communication_method = models.IntegerField(
        choices=CommunicationMethod.choices())


class StaffingResponse(StaffingResponseMixin, BaseModel):
    """
    A StaffingResponse object stores a response from a worker and
    whether the worker got the task.

    Attributes:
        request_inquiry (orchestra.models.StaffingRequestInquiry):
            Request inquiry associated with a response
        response_text (str):
            Response text that a Worker provided
        is_available (bool):
            True if a Worker is ready to work on the Task
        is_winner (bool):
            True if a Worker was selected to work on the Task
    """
    request_inquiry = models.ForeignKey(StaffingRequestInquiry,
                                        on_delete=models.CASCADE,
                                        related_name='responses')
    response_text = models.TextField(blank=True, null=True)
    is_available = models.BooleanField()
    is_winner = models.NullBooleanField()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0008_auto_20150520_1953'),
    ]

    operations = [
        migrations.AlterField(
            model_name='certification',
            name='required_certifications',
            field=models.ManyToManyField(
                blank=True, to='orchestra.Certification', related_name='required_certifications_rel_+'),
        ),
        migrations.AlterField(
            model_name='task',
            name='step_slug',
            field=models.CharField(max_length=200, choices=[
                                   ('content_extraction', ' Content Extraction'), ('copy_pass', 'Copy Pass')]),
        ),
        migrations.AlterField(
            model_name='worker',
            name='user',
            field=models.OneToOneField(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]



from orchestra.utils.json_schema import DefaultValidatingDraft4Validator


def validate_json(blob_key, json_schema, data):
    """
    Validate json blobs against their json schema.
    :param blob_key: key name for data in model
    :param json_schema: JSON schema object
    :param data: JSON data, defaults are added to it.
    """
    if not data:
        # Don't validate when data is empty. Otherwise, won't be able to
        # create empty database objects.
        return data
    schema = json_schema.get_schema()
    DefaultValidatingDraft4Validator(schema).validate(data)
    return data



import logging
from pydoc import locate

from django.db import transaction
from django.utils import timezone

from orchestra.core.errors import MachineExecutionError
from orchestra.models import Iteration
from orchestra.models import Project
from orchestra.models import Step
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.utils.task_lifecycle import create_subsequent_tasks
from orchestra.utils.task_lifecycle import get_previously_completed_task_data
from orchestra.utils.task_properties import get_latest_iteration

logger = logging.getLogger(__name__)


def execute(project_id, step_slug):
    project = Project.objects.get(id=project_id)
    step = Step.objects.get(slug=step_slug,
                            workflow_version=project.workflow_version)
    task = Task.objects.get(project=project,
                            step=step)

    # Run machine function
    if step.is_human:
        raise MachineExecutionError('Step worker type is not machine')

    if task.status == Task.Status.COMPLETE:
        raise MachineExecutionError('Task assignment already completed')

    # Machine tasks are only assigned to one worker/machine,
    # so they should only have one task assignment,
    # and should never be submitted for review.

    with transaction.atomic():
        # Uniqueness constraint on assignnment_counter and task prevents
        # concurrent creation of more than one assignment
        task_assignment, created = TaskAssignment.objects.get_or_create(
            assignment_counter=0,
            task=task,
            defaults={
                'status': TaskAssignment.Status.PROCESSING,
                'in_progress_task_data': {}})
        if created:
            task.status = Task.Status.PROCESSING
            task.save()

            Iteration.objects.create(
                assignment=task_assignment,
                start_datetime=task_assignment.start_datetime)
        else:
            # Task assignment already exists
            if task_assignment.status == TaskAssignment.Status.FAILED:
                # Pick up failed task for reprocessing
                task_assignment.status = TaskAssignment.Status.PROCESSING
                task_assignment.save()
            else:
                # Task assignment already processing
                raise MachineExecutionError(
                    'Task already picked up by another machine')

    prerequisites = get_previously_completed_task_data(task.step, task.project)

    function = locate(step.execution_function['path'])
    kwargs = step.execution_function.get('kwargs', {})
    try:
        project_data = project.project_data
        project_data['project_id'] = project_id
        task_data = function(project_data, prerequisites, **kwargs)
    except Exception:
        task_assignment.status = TaskAssignment.Status.FAILED
        logger.exception('Machine task has failed')
        task_assignment.save()
        return
    task_assignment.status = TaskAssignment.Status.SUBMITTED
    task_assignment.in_progress_task_data = task_data
    task_assignment.save()

    if task.project.status == Project.Status.ABORTED:
        # If a long-running task's project was aborted while running, we ensure
        # the aborted state on the task.
        task.status = Task.Status.ABORTED
        task.save()
    else:
        task.status = Task.Status.COMPLETE
        task.save()

        iteration = get_latest_iteration(task_assignment)
        iteration.status = Iteration.Status.REQUESTED_REVIEW
        iteration.submitted_data = task_data
        iteration.end_datetime = timezone.now()
        iteration.save()

        create_subsequent_tasks(project)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-05-20 03:09
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0063_worker_staffing_priority'),
    ]

    operations = [
        migrations.AddField(
            model_name='workercertification',
            name='staffbot_enabled',
            field=models.BooleanField(default=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-04-04 21:09
from __future__ import unicode_literals

import django.db.models.deletion
import django.utils.timezone
import jsonfield.fields
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0024_auto_20160325_1916'),
    ]

    operations = [
        migrations.CreateModel(
            name='Iteration',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('start_datetime', models.DateTimeField(
                    default=django.utils.timezone.now)),
                ('end_datetime', models.DateTimeField(blank=True, null=True)),
                ('status', models.IntegerField(choices=[
                 (0, 'Processing'), (1, 'Requested Review'), (2, 'Provided Review')])),
                ('submitted_data', jsonfield.fields.JSONField(blank=True, default={})),
                ('assignment', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                                 related_name='iterations', to='orchestra.TaskAssignment')),
            ],
        ),
    ]



from django.db.models import Max
from django.db.models import Q
from django.utils import timezone
from pydoc import locate

from orchestra.core.errors import SanityBotError
from orchestra.models import Project
from orchestra.models import SanityCheck
from orchestra.models import WorkflowVersion
from orchestra.utils.notifications import message_experts_slack_group


def _handle(project, sanity_check, handler):
    handler_type = handler.get('type')
    handler_message = handler.get('message')
    handler_steps = handler.get('steps')
    if (handler_type != 'slack_project_channel' or
            not handler_message or not handler_steps):
        raise SanityBotError('Invalid handler: {}'.format(handler))
    tasks = (
        task for task in project.tasks.all()
        if task.step.slug in handler_steps)
    usernames = {
        assignment.worker.formatted_slack_username()
        for task in tasks
        for assignment in task.assignments.all()
        if assignment and assignment.worker
    }
    message = '{}: {}'.format(' '.join(usernames), handler_message)
    message_experts_slack_group(
        project.slack_group_id, message)


def _filter_checks(project, checks, check_configurations):
    latest_check_creation = {
        check['check_slug']: check['max_created_at']
        for check in (SanityCheck.objects
                      .filter(project=project)
                      .values('check_slug')
                      .annotate(max_created_at=Max('created_at')))}
    for check in checks:
        max_created_at = latest_check_creation.get(check.check_slug)
        seconds = (
            check_configurations.get(check.check_slug, {})
            .get('repetition_seconds'))
        now = timezone.now()
        if (max_created_at is None or
                ((seconds is not None) and
                 ((now - max_created_at).total_seconds() > seconds))):
            yield check


def _handle_sanity_checks(project, sanity_checks, check_configurations):
    sanity_checks = _filter_checks(
        project, sanity_checks, check_configurations)
    for sanity_check in sanity_checks:
        config = check_configurations.get(sanity_check.check_slug)
        if config is None:
            raise SanityBotError(
                'No configuration for {}'.format(sanity_check.check_slug))
        handlers = config.get('handlers')
        if handlers is None:
            raise SanityBotError(
                'No handlers for {}'.format(sanity_check.check_slug))
        for handler in handlers:
            _handle(project, sanity_check, handler)
        sanity_check.handled_at = timezone.now()
        sanity_check.project = project
        sanity_check.save()


def create_and_handle_sanity_checks():
    workflow_versions = WorkflowVersion.objects.all()
    incomplete_projects = (Project.objects
                           .filter(workflow_version__in=workflow_versions)
                           .filter(Q(status=Project.Status.ACTIVE) |
                                   Q(status=Project.Status.PAUSED)))

    for project in incomplete_projects:
        sanity_checks = project.workflow_version.sanity_checks
        sanity_check_path = (sanity_checks
                             .get('sanity_check_function', {})
                             .get('path'))
        check_configurations = sanity_checks.get('check_configurations')
        if sanity_check_path and check_configurations:
            sanity_check_function = locate(sanity_check_path)
            sanity_checks = sanity_check_function(project)
            _handle_sanity_checks(
                project, sanity_checks, check_configurations)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0011_auto_20150618_0003'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='process_slug',
            field=models.CharField(max_length=200, choices=[(
                'website_design', 'Website design'), ('website_enhancement', 'Website enhancement')]),
        ),
        migrations.AlterField(
            model_name='project',
            name='short_description',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='task',
            name='step_slug',
            field=models.CharField(max_length=200, choices=[('design', 'Design'), ('export', 'Export design'), (
                'website_enhancement', 'Enhance'), ('export_enhancement_step', 'Export enhancement')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-29 16:42
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0052_auto_20160428_2142'),
    ]

    operations = [
        migrations.AddField(
            model_name='worker',
            name='slack_user_id',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
    ]



import jsl

TAG_STATUS = [
    'default',
    'primary',
    'success',
    'info',
    'warning',
    'danger',
]


class TagSchema(jsl.Document):
    """
    A tag schema

    Attributes:
        label (str):
            A text label of the tag
        status (str):
            The tag status, limited to one of the available
            string in TAG_STATUS
    """
    label = jsl.StringField(required=True)
    status = jsl.StringField(enum=TAG_STATUS, default='default')


class TagListSchema(jsl.Document):
    tags = jsl.ArrayField(jsl.DocumentField('TagSchema'))



from unittest.mock import patch

from django.conf import settings
from django.test import override_settings

from orchestra.communication.mail import send_mail
from orchestra.models import CommunicationPreference
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import setup_models


class ModelsTestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = self.workers[0]
        self.comm_type = (CommunicationPreference.CommunicationType
                          .TASK_STATUS_CHANGE.value)
        self.recipient_list = [self.worker.user.email]

    @patch('orchestra.communication.mail._send_mail')
    def test_filtering_no_preference(self, mock_mail):
        """
            Verify that we correctly filter users based on their email
            preferences.
        """
        # Test when no comm_type is given
        send_mail(subject='test_subject',
                  message='test_message',
                  from_email='test@test.com',
                  recipient_list=self.recipient_list
                  )
        mock_mail.assert_called_once_with(
            'test_subject', 'test_message',
            'test@test.com', self.recipient_list,
            fail_silently=False, auth_user=None,
            auth_password=None, connection=None,
            html_message=None
        )

    @override_settings(ORCHESTRA_MOCK_TO_EMAIL='to-email@test.com')
    @override_settings(ORCHESTRA_MOCK_EMAILS=True)
    @patch('orchestra.communication.mail._send_mail')
    def test_email_mocked(self, mock_mail):
        """
            Verify that we mock email sending.
        """
        # Test when no comm_type is given
        send_mail(subject='test_subject',
                  message='test_message',
                  from_email='test@test.com',
                  recipient_list=self.recipient_list,
                  )
        mock_mail.assert_called_once_with(
            'test_subject', 'test_message',
            'test@test.com', [settings.ORCHESTRA_MOCK_TO_EMAIL],
            fail_silently=False, auth_user=None,
            auth_password=None, connection=None,
            html_message=None
        )

    @patch('orchestra.communication.mail._send_mail')
    def test_filtering_send_with_preference(self, mock_mail):
        # Test when comm_type is allowed
        send_mail(subject='test_subject',
                  message='test_message',
                  from_email='test@test.com',
                  recipient_list=self.recipient_list,
                  communication_type=self.comm_type
                  )
        mock_mail.assert_called_once_with(
            'test_subject', 'test_message',
            'test@test.com', self.recipient_list,
            fail_silently=False, auth_user=None,
            auth_password=None, connection=None,
            html_message=None
        )

    @patch('orchestra.communication.mail._send_mail')
    def test_filtering_no_send(self, mock_mail):
        # Test when comm_type is not allowed
        comm_pref = CommunicationPreference.objects.get(
            worker=self.worker,
            communication_type=self.comm_type
        )
        comm_pref.methods.email = ~CommunicationPreference.methods.email
        comm_pref.save()
        self.assertFalse(comm_pref.can_email())
        send_mail(subject='test_subject',
                  message='test_message',
                  from_email='test@test.com',
                  recipient_list=self.recipient_list,
                  communication_type=self.comm_type
                  )
        self.assertFalse(mock_mail.called)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2018-01-16 15:23
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0073_remove_worker_staffing_priority'),
    ]

    operations = [
        migrations.AddField(
                model_name='todo',
                name='start_by_datetime',
                field=models.DateTimeField(blank=True, null=True),
            ),
        migrations.AddField(
            model_name='todo',
            name='due_datetime',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



import logging

from django.http import JsonResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from orchestra.bots.basebot import BaseBot
from orchestra.bots.errors import SlackCommandInvalidRequest
from orchestra.bots.errors import SlackUserUnauthorized
from orchestra.bots.staffbot import StaffBot
from orchestra.communication.slack import format_slack_message

logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name='dispatch')
class BotMixin(View):
    """
    Generic mixin to handle messages to bots, should be used by specifying
    a `BotClass` to instantiate with the request data.
    """
    BotClass = BaseBot
    bot_config = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bot = self.BotClass(**self.bot_config)

    def post(self, request, *args, **kwargs):
        data = request.POST
        try:
            response_data = self.bot.dispatch(data)
        except SlackCommandInvalidRequest as e:
            logger.exception('Invalid slack command')
            response_data = format_slack_message(str(e))
        except SlackUserUnauthorized as e:
            logger.exception('Unauthorized slack user')
            response_data = format_slack_message(str(e))

        return JsonResponse(response_data)


class StaffBotView(BotMixin):
    BotClass = StaffBot



class MockMail:
    def __init__(self):
        self.inbox = []

    def send_mail(self, recipient_list, **kwargs):
        for recipient in recipient_list:
            self.inbox.append({
                'recipient': recipient,
                'subject': kwargs['subject'],
                'message': kwargs['message']
            })

    def clear(self):
        self.inbox[:] = []



from rest_framework import serializers

from orchestra.models import Todo
from orchestra.models import TodoQA
from orchestra.models import TodoListTemplate
from orchestra.json_schemas.todos import TodoListSchema
from orchestra.json_schemas.todos import TodoActionListSchema
from orchestra.utils.mixins import JSONSchemaValidationMixin


class TodoQASerializer(serializers.ModelSerializer):

    class Meta:
        model = TodoQA
        fields = (
            'id',
            'created_at',
            'todo',
            'approved',
            'comment')
        read_only_fields = ('id',)


class TodoSerializer(serializers.ModelSerializer):
    qa = serializers.SerializerMethodField()
    json_schemas = {
        'activity_log': TodoActionListSchema
    }

    def get_qa(self, obj):
        return None

    class Meta:
        model = Todo
        fields = (
            'id',
            'created_at',
            'task',
            'description',
            'parent_todo',
            'template',
            'qa',
            'completed',
            'skipped_datetime',
            'start_by_datetime',
            'due_datetime',
            'activity_log')
        read_only_fields = ('id',)


class TodoWithQASerializer(serializers.ModelSerializer):
    qa = TodoQASerializer(read_only=True)
    json_schemas = {
        'activity_log': TodoActionListSchema
    }

    class Meta:
        model = Todo
        fields = (
            'id',
            'created_at',
            'task',
            'description',
            'parent_todo',
            'template',
            'qa',
            'completed',
            'skipped_datetime',
            'start_by_datetime',
            'due_datetime',
            'activity_log')
        read_only_fields = ('id',)


class TodoListTemplateSerializer(serializers.ModelSerializer,
                                 JSONSchemaValidationMixin):
    json_schemas = {
        'todos': TodoListSchema
    }

    class Meta:
        model = TodoListTemplate
        fields = (
            'id',
            'created_at',
            'slug',
            'name',
            'description',
            'creator',
            'todos')
        read_only_fields = ('id',)



from orchestra.models import Project


def completed_projects(projects):
    """
    Filters `projects` queryset to completed ones.
    """
    return projects.filter(status=Project.Status.COMPLETED)



import json

from orchestra.tests.helpers import OrchestraTestCase
from orchestra.utils.load_json import load_encoded_json


class LoadEncodedJsonTest(OrchestraTestCase):

    def test_non_json_string(self):
        encoded_data = ''
        result = load_encoded_json(encoded_data)
        expected = {}
        self.assertEqual(result, expected)

    def test_decode_error(self):
        encoded_data = 1
        result = load_encoded_json(encoded_data)
        expected = {}
        self.assertEqual(result, expected)

    def test_valid(self):
        expected = {'test_key': 'test_value'}
        encoded_data = json.dumps(expected).encode()
        result = load_encoded_json(encoded_data)
        self.assertEqual(result, expected)





def get_mock_slack_data(
        token='test-token',
        team_id='test-team-id',
        team_domain='test-team-domain',
        channel_id='test-channel-id',
        channel_name='test-channel-name',
        user_id='test-user-id',
        user_name='test-user-name',
        command='/test',
        text='test-text'):
    return {
        'token': token,
        'team_id': team_id,
        'team_domain': team_domain,
        'channel_id': channel_id,
        'channel_name': channel_name,
        'user_id': user_id,
        'user_name': user_name,
        'command': command,
        'text': text,
    }



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2018-01-21 15:41
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import jsonfield.fields
import orchestra.models.core.mixins
import orchestra.utils.models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0074_auto_20180116_1523'),
    ]

    operations = [
        migrations.CreateModel(
            name='SanityCheck',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now)),
                ('is_deleted', models.BooleanField(default=False)),
                ('handled_at', models.DateTimeField(blank=True, null=True)),
                ('check_slug', models.CharField(max_length=200)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sanity_checks', to='orchestra.Project')),
            ],
            bases=(orchestra.models.core.mixins.SanityCheckMixin, orchestra.utils.models.DeleteMixin, models.Model),
        ),
        migrations.AddField(
            model_name='workflowversion',
            name='sanity_checks',
            field=jsonfield.fields.JSONField(default={}),
        ),
    ]



# -*- coding: utf-8 -*-
# Manually written
from __future__ import unicode_literals

import logging

from django.db import connection
from django.db import migrations
from django.db import models

logger = logging.getLogger(__name__)


def delete_timeentries_marked_for_delete(apps, schema_editor):
    TimeEntry = apps.get_model('orchestra', 'TimeEntry')
    TimeEntry.objects.filter(is_deleted=True).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0050_auto_20160428_1751'),
    ]

    operations = [
        migrations.RunPython(delete_timeentries_marked_for_delete),  # manually-reviewed
    ]



import datetime

from dateutil.parser import parse

from orchestra.accounts.signals import orchestra_user_registered
from orchestra.core.errors import ModelSaveError
from orchestra.models import PayRate
from orchestra.models import Worker
from orchestra.models import WorkerCertification
from orchestra.tests.helpers import OrchestraModelTestCase
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import CertificationFactory
from orchestra.tests.helpers.fixtures import PayRateFactory
from orchestra.tests.helpers.fixtures import ProjectFactory
from orchestra.tests.helpers.fixtures import StepFactory
from orchestra.tests.helpers.fixtures import TaskAssignmentFactory
from orchestra.tests.helpers.fixtures import TaskFactory
from orchestra.tests.helpers.fixtures import UserFactory
from orchestra.tests.helpers.fixtures import WorkerCertificationFactory
from orchestra.tests.helpers.fixtures import WorkerFactory
from orchestra.tests.helpers.fixtures import WorkflowFactory
from orchestra.tests.helpers.fixtures import WorkflowVersionFactory
from orchestra.tests.helpers.fixtures import setup_models


class ModelsTestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.workers[0] = Worker.objects.get(user__username='test_user_1')
        self.worker = self.workers[0]

    def test_certification_roles(self):
        """ Ensure that workers can be certified at multiple roles. """
        certification = CertificationFactory(
            slug='cat_herding', name='Cat herding',
            workflow=self.workflows['w1'])
        WorkerCertificationFactory(
            certification=certification,
            worker=self.workers[0],
            role=WorkerCertification.Role.ENTRY_LEVEL)
        WorkerCertificationFactory(
            certification=certification,
            worker=self.workers[0],
            role=WorkerCertification.Role.REVIEWER)

    def test_reviewer_requires_entry_level(self):
        """ Workers must have entry-level certification before they review. """
        certification = CertificationFactory(
            slug='cat_herding', name='Cat herding',
            workflow=self.workflows['w1'])
        with self.assertRaises(ModelSaveError):
            WorkerCertificationFactory(
                certification=certification,
                worker=self.workers[0],
                role=WorkerCertification.Role.REVIEWER)

    def test_worker_autocreation(self):
        """ When new users register, worker objects should be autocreated. """

        # Create a user
        user = UserFactory(username='test_registration_user',
                           password='test',
                           email='test_registration_user@test.com')

        # There should be no worker yet
        self.assertFalse(Worker.objects.filter(user=user).exists())

        # Fake registering the user
        orchestra_user_registered.send(
            sender=self.__class__, user=user, request=None)

        # Expect the worker object to be created
        self.assertTrue(Worker.objects.filter(user=user).exists(),
                        'Worker not autocreated on User registration')

    def test_pay_rate_creation(self):
        """
        Ensure that PayRates are created with valid ranges and do not
        overlap with existing entries.
        """
        start_stop_dates = [
            (('2016-01-01', '2016-01-03'), ('2016-01-02', None), False),
            (('2016-01-01', None), ('2016-02-01', None), False),
            (('2016-01-01', None), ('2015-12-01', '2016-01-01'), False),
            (('2016-01-01', '2016-01-10'), ('2016-01-03', '2016-01-08'),
             False),
            (('2016-01-01', '2016-01-10'), ('2015-12-30', '2016-01-03'),
             False),
            (('2016-01-01', '2016-01-10'), ('2016-01-03', '2016-01-18'),
             False),
            (('2016-01-01', '2016-01-10'), ('2015-12-30', '2016-01-18'),
             False),
            (('2016-01-01', '2016-01-10'), ('2016-01-30', None), True),
            (('2016-01-01', '2016-01-10'), ('2016-01-30', '2016-02-18'), True),
            (('2016-03-02', None), ('2016-01-01', '2016-03-01'), True)
        ]

        def _parse_date(date):
            if date:
                start_date = parse(date).date()
            else:
                start_date = None
            return start_date

        def _verify_failure(saved, new, success):
            start_date_saved = _parse_date(saved[0])
            end_date_saved = _parse_date(saved[1])
            PayRate.objects.create(worker=self.worker,
                                   hourly_rate=10,
                                   hourly_multiplier=1.2,
                                   start_date=start_date_saved,
                                   end_date=end_date_saved)
            start_date_new = _parse_date(new[0])
            end_date_new = _parse_date(new[1])
            payrate = PayRate(worker=self.worker,
                              hourly_rate=10,
                              hourly_multiplier=1.2,
                              start_date=start_date_new,
                              end_date=end_date_new)
            if success:
                payrate.save()
            else:
                with self.assertRaises(ModelSaveError):
                    payrate.save()
            PayRate.objects.all().delete()

        for saved, new, success in start_stop_dates:
            _verify_failure(saved, new, success)

    def test_pay_rate_update(self):
        """ Verify that we don't detect overlap collision with itself. """
        payrate = PayRate.objects.create(
            worker=self.worker,
            hourly_rate=10,
            hourly_multiplier=1.2,
            start_date=datetime.date(2016, 1, 1),
            end_date=None)
        payrate.end_date = datetime.date(2016, 3, 1)
        payrate.save()


class CertificationTestCase(OrchestraModelTestCase):
    __test__ = True
    model = CertificationFactory


class StepTestCase(OrchestraModelTestCase):
    __test__ = True
    model = StepFactory


class PayRateTestCase(OrchestraModelTestCase):
    __test__ = True
    model = PayRateFactory


class ProjectTestCase(OrchestraModelTestCase):
    __test__ = True
    model = ProjectFactory


class TaskTestCase(OrchestraModelTestCase):
    __test__ = True
    model = TaskFactory

    def test_get_detailed_description(self):
        """
        Verify that the detailed description text is valid
        """
        # description functions are optional
        task = TaskFactory()
        self.assertEqual(task.get_detailed_description(), '')

        no_kwargs = {
            'path': ('orchestra.tests.helpers.'
                     'fixtures.get_detailed_description')
        }
        task_no_kwargs = TaskFactory(
            step=StepFactory(slug='stepslug',
                             detailed_description_function=no_kwargs))
        self.assertEqual(task_no_kwargs.get_detailed_description(),
                         'No text given stepslug')

        with_kwargs = {
            'path': ('orchestra.tests.helpers.'
                     'fixtures.get_detailed_description'),
            'kwargs': {
                'text': 'task 2 text',
            }
        }
        task_with_kwargs = TaskFactory(
            step=StepFactory(slug='stepslug',
                             detailed_description_function=with_kwargs))
        self.assertEqual(task_with_kwargs.get_detailed_description(),
                         'task 2 text stepslug')

        extra_kwargs = {'text': 'extra text'}
        self.assertEqual(
            task_with_kwargs.get_detailed_description(
                extra_kwargs=extra_kwargs),
            'extra text stepslug'
        )


class TaskAssignmentTestCase(OrchestraModelTestCase):
    __test__ = True
    model = TaskAssignmentFactory


class WorkerTestCase(OrchestraModelTestCase):
    __test__ = True
    model = WorkerFactory

    def setUp(self):
        self.cert = CertificationFactory()
        super().setUp()

    def test_is_reviewer(self):
        worker = WorkerFactory()
        WorkerCertificationFactory(
            worker=worker,
            certification=self.cert,
            role=WorkerCertification.Role.ENTRY_LEVEL)
        WorkerCertificationFactory(
            worker=worker,
            certification=self.cert,
            role=WorkerCertification.Role.REVIEWER)
        self.assertTrue(worker.is_reviewer(self.cert))
        self.assertTrue(worker.is_entry_level(self.cert))

    def test_is_entry_level(self):
        worker = WorkerFactory()
        WorkerCertificationFactory(
            worker=worker,
            certification=self.cert,
            role=WorkerCertification.Role.ENTRY_LEVEL)
        self.assertTrue(worker.is_entry_level(self.cert))
        self.assertFalse(worker.is_reviewer(self.cert))


class WorkerCertificationTestCase(OrchestraModelTestCase):
    __test__ = True
    model = WorkerCertificationFactory


class WorkflowTestCase(OrchestraModelTestCase):
    __test__ = True
    model = WorkflowFactory


class WorkflowVersionTestCase(OrchestraModelTestCase):
    __test__ = True
    model = WorkflowVersionFactory
